/*      $OpenBSD: ex_undo.c,v 1.6 2014/11/12 04:28:41 bentley Exp $     */

/* SPDX-License-Identifier: BSD-3-Clause */

/*
 * Copyright (c) 1992, 1993, 1994
 *      The Regents of the University of California.  All rights reserved.
 * Copyright (c) 1992, 1993, 1994, 1995, 1996
 *      Keith Bostic.  All rights reserved.
 * Copyright (c) 2022-2024 Jeffrey H. Johnson <trnsz@pobox.com>
 *
 * See the LICENSE.md file for redistribution information.
 */

#include <sys/types.h>
#include <sys/queue.h>

#include <bitstring.h>
#include <limits.h>
#include <stdio.h>
#include <bsd_stdlib.h>

#include "../common/common.h"

/*
 * ex_undo -- u
 *      Undo the last change.
 *
 * PUBLIC: int ex_undo(SCR *, EXCMD *);
 */
int
ex_undo(SCR *sp, EXCMD *cmdp)
{
        EXF *ep;
        MARK m;

        /*
         * !!!
         * Historic undo always set the previous context mark.
         */
        m.lno = sp->lno;
        m.cno = sp->cno;
        if (mark_set(sp, ABSMARK1, &m, 1))
                return (1);

        /*
         * !!!
         * Multiple undo isn't available in ex, as there's no '.' command.
         * Whether 'u' is undo or redo is toggled each time, unless there
         * was a change since the last undo, in which case it's an undo.
         */
        ep = sp->ep;
        if (!F_ISSET(ep, F_UNDO)) {
                F_SET(ep, F_UNDO);
                ep->lundo = FORWARD;
        }
        switch (ep->lundo) {
        case BACKWARD:
                if (log_forward(sp, &m))
                        return (1);
                ep->lundo = FORWARD;
                break;
        case FORWARD:
                if (log_backward(sp, &m))
                        return (1);
                ep->lundo = BACKWARD;
                break;
        case NOTSET:
                abort();
        }
        sp->lno = m.lno;
        sp->cno = m.cno;
        return (0);
}
